from django.apps import AppConfig
from django.conf import settings


from django.db.models.signals import post_delete, post_save


class IDPUserConfig(AppConfig):
    name = "idp_user"

    def ready(self):
        # If Kafka is not configured, do not register signals
        is_kafka_configured = getattr(settings, "KAFKA_ARN", None) or getattr(
            settings, "KAFKA_BROKER", None
        )
        if not is_kafka_configured:
            return

        from idp_user.services.base_user import BaseUserService
        from idp_user.settings import APP_ENTITIES

        for (
            _app_entity_type,
            config,
        ) in APP_ENTITIES.items():
            model = config["model"]
            post_save.connect(
                receiver=BaseUserService.process_app_entity_record_post_save,
                sender=model,
            )
            post_delete.connect(
                receiver=BaseUserService.process_app_entity_record_post_delete,
                sender=model,
            )



# Generated by Django 4.0.6 on 2022-08-17 15:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('idp_user', '0003_auto_20220513_1025'),
    ]

    operations = [
        migrations.RenameField(
            model_name='user',
            old_name='idp_user_id',
            new_name='id',
        ),
        migrations.AlterField(
            model_name='user',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



from typing import Optional

import requests
from django.conf import settings
from django.contrib.auth.backends import ModelBackend

from idp_user.models.user import User
from idp_user.utils.functions import get_or_none

IDP_URL = settings.IDP_USER_APP.get("IDP_URL")
HTTP_200_OK = 200


class IDPAuthBackend(ModelBackend):
    def authenticate(self, request, **kwargs):
        access_token = self._fetch_token(request)
        if not access_token:
            return None

        response = requests.get(
            url=f"{IDP_URL}/api/users/me/",
            headers={"Authorization": f"Bearer {access_token}"},
        )
        username = response.json()["username"]
        return get_or_none(User.objects, username=username)

    @staticmethod
    def _fetch_token(request) -> Optional[str]:
        response = requests.post(
            url=f"{IDP_URL}/api/login/",
            json={
                "username": request.POST.get("username"),
                "password": request.POST.get("password"),
            },
        )

        if response.status_code == HTTP_200_OK:
            return response.json()["access"]



import faust
from asgiref.sync import sync_to_async
from django.conf import settings
from django.utils.module_loading import import_string
from faust import StreamT

from idp_user.services import UserService
from idp_user.settings import CONSUMER_APP_ENV

app = import_string(settings.IDP_USER_APP["FAUST_APP_PATH"])

"""
{
    "username": "",
    "first_name": "",
    "last_name": "",
    "email": "",
    "app_specific_configs": {
        "app_identifier": {
            "tenant": {
                "Servicer": {
                    "app_entities_restrictions": {"vehicle": [1, 2]},
                    "permission_restrictions": {
                        "synchronizeDoD": False
                    }
                }
            },
        }
    }
}
"""


class UserRecord(faust.Record):
    first_name: str
    last_name: str
    username: str
    email: str
    is_active: bool
    is_staff: bool
    is_superuser: bool
    date_joined: str
    app_specific_configs: dict


USER_UPDATES_TOPIC_NAME = f"{CONSUMER_APP_ENV}_user_updates"

user_updates = app.topic(USER_UPDATES_TOPIC_NAME, value_type=UserRecord)


async def update_user(user_record: UserRecord):
    await sync_to_async(UserService.process_user)(user_record.asdict())


async def verify_if_user_exists_and_delete_roles(user_record: UserRecord):
    await sync_to_async(UserService.verify_if_user_exists_and_delete_roles)(
        user_record.asdict()
    )


@app.agent(user_updates)
async def update_user_stream_processor(user_records: StreamT[UserRecord]):
    async for user_record in user_records:
        if user_record.app_specific_configs.get(
            settings.IDP_USER_APP["APP_IDENTIFIER"]
        ):
            await update_user(user_record)
        else:
            # Having arrived here means that the user does not have access in the current app
            # Verify however if the user already exists in the database of any tenant
            # If this is the case, delete his/her roles
            await verify_if_user_exists_and_delete_roles(user_record)



class ChoicesMixin:
    """
    A helper class, mostly useful to declare roles, in the form:

    class Roles(ChoicesMixin):
        RoleName = "role_database_name"
    """

    @classmethod
    def choices(cls):
        return {
            label: value
            for (label, value) in cls.__dict__.items()
            if not label.startswith("_")
        }

    @classmethod
    def as_list(cls):
        return [
            (value, label)
            for (label, value) in cls.__dict__.items()
            if not label.startswith("_")
        ]

    @classmethod
    def as_dict(cls):
        return {
            value: label
            for (label, value) in cls.__dict__.items()
            if not label.startswith("_")
        }



import logging
from collections import defaultdict
from copy import deepcopy
from typing import Any, Optional, Union

from django.conf import settings
from django.core.cache import cache
from django.db import models, transaction
from django.db.models import Q, QuerySet
from rest_framework.exceptions import PermissionDenied
from rest_framework.request import Request

from idp_user.models import UserRole
from idp_user.models.user import User
from idp_user.services.base_user import BaseUserService
from idp_user.settings import APP_ENTITIES, APP_IDENTIFIER, IN_DEV, ROLES, TENANTS
from idp_user.signals import (
    post_create_idp_user,
    post_update_idp_user,
    pre_update_idp_user,
)
from idp_user.utils.functions import (
    cache_user_service_results,
    get_or_none,
    keep_keys,
    update_record,
)
from idp_user.utils.typing import (
    ALL,
    AppEntityTypeConfig,
    UserRecordDict,
    UserTenantData,
)

logger = logging.getLogger(__name__)


class UserService(BaseUserService):
    # Service Methods Used by Django Application
    @staticmethod
    def get_role(request: Request):
        return request.query_params.get("role")

    @staticmethod
    def authorize_and_get_records_or_get_all_allowed(
        user: User,
        role: ROLES,
        app_entity_type: str,
        app_entity_records_identifiers: Optional[list[Any]],
        permission: str = None,
    ) -> models.QuerySet:
        """
        Make sure that the user can access the entity records being requested and return them.
        If no identifiers are being provided, return all allowed entity records.

        Args:
            user:                           The user performing the request
            role:                           The role that the user is acting as.
            app_entity_type:                The app entity being accessed
            app_entity_records_identifiers: The identifiers of the records belonging to the specified app entity
            permission:                     In case of specific permissions we can have permission restrictions
                                                through IDP. The value is the name of the permission


        Raises:
            PermissionDenied: In case the requested records are not allowed
        """

        if not app_entity_records_identifiers:
            return UserService.get_allowed_app_entity_records(
                user=user,
                role=role,
                app_entity_type=app_entity_type,
                permission=permission,
            )
        UserService.authorize_app_entity_records(
            user=user,
            role=role,
            app_entity_type=app_entity_type,
            app_entity_records_identifiers=app_entity_records_identifiers,
            permission=permission,
        )
        return UserService._get_records(
            app_entity_type=app_entity_type,
            records_identifiers=app_entity_records_identifiers,
        )

    @staticmethod
    def authorize_app_entity_records(
        user: User,
        role: ROLES,
        app_entity_type: str,
        app_entity_records_identifiers: list[Any],
        permission: str = None,
    ):
        """
        Verify if the user has access to the requested entity records.
        If a permission is specified and it has restrictable=True, the access is verified based on it.

        Args:
            user:                           The user performing the request
            role:                           The role that the user is acting as.
            app_entity_type:                The app entity being accessed
            app_entity_records_identifiers: The identifiers of the records belonging to the specified app entity
            permission:                     In case of specific permissions we can have permission restrictions
                                                through IDP. The value is the name of the permission

        Raises:
            PermissionDenied: In case the requested records are not allowed
        """

        assert (
            app_entity_type in APP_ENTITIES.keys()
        ), f"Unknown app entity: {app_entity_type}!"

        allowed_app_entity_records_identifiers = (
            UserService._get_allowed_app_entity_records_identifiers(
                user=user,
                role=role,
                app_entity_type=app_entity_type,
                permission=permission,
            )
        )

        if allowed_app_entity_records_identifiers == ALL:
            return

        if not set(app_entity_records_identifiers).issubset(
            set(allowed_app_entity_records_identifiers)
        ):
            raise PermissionDenied(
                "You are not allowed to access the records in the requested entity!"
            )

    @staticmethod
    def get_allowed_app_entity_records(
        user: User, role: ROLES, app_entity_type: str, permission: str = None
    ) -> models.QuerySet:
        """
        Gets the app entity records that the user can access.

        Args:
            user:               The user performing the request
            role:               The role that the user is acting as.
            app_entity_type:    The app entity being accessed
            permission:         In case of specific permissions we can have permission restrictions
                                    through IDP. The value is the name of the permission

        Returns:
            QuerySet of App Entity Records that the user can access
        """

        assert (
            app_entity_type in APP_ENTITIES.keys()
        ), f"Unknown app entity: {app_entity_type}!"

        allowed_app_entity_records_identifiers = (
            UserService._get_allowed_app_entity_records_identifiers(
                user=user,
                role=role,
                app_entity_type=app_entity_type,
                permission=permission,
            )
        )

        return UserService._get_records(
            app_entity_type=app_entity_type,
            records_identifiers=allowed_app_entity_records_identifiers,
        )

    @staticmethod
    def _get_app_entity_type_configs(app_entity_type: str) -> AppEntityTypeConfig:
        try:
            return APP_ENTITIES[app_entity_type]
        except KeyError:
            raise KeyError(
                f"No config declared for app_entity_type={app_entity_type} "
                f"in IDP_USER_APP['APP_ENTITIES']!"
            ) from None

    @staticmethod
    def _get_records(
        app_entity_type: str, records_identifiers: Union[list[Any], ALL]
    ) -> models.QuerySet:
        app_entity_type_configs = UserService._get_app_entity_type_configs(
            app_entity_type
        )
        model = app_entity_type_configs["model"]

        if records_identifiers == ALL:
            return model.objects.all()
        else:
            model_identifier_attr = app_entity_type_configs["identifier_attr"]
            return model.objects.filter(
                **{f"{model_identifier_attr}__in": records_identifiers}
            )

    @staticmethod
    @cache_user_service_results
    def _get_allowed_app_entity_records_identifiers(
        user: User, role: ROLES, app_entity_type: str, permission: str = None
    ) -> Union[list[Any], ALL]:
        """
        Gets the identifiers of the app entity records that the user can access.
        If no restriction both on role and permission level, return '__all__'

        Args:
            user:               The user performing the request
            role:               The role that the user is acting as.
            app_entity_type:    The app entity being accessed
            permission:         In case of specific permissions we can have permission restrictions
                                    through IDP. The value is the name of the permission

        Returns:
            List of identifiers of App Entity Records that the user can access
        """
        assert ROLES.as_dict().get(role) is not None, f"Role does not exist: {role}"

        try:
            user_role = UserRole.objects.get(user=user, role=role)
        except UserRole.DoesNotExist:
            return []

        # Permission restriction get precedence, if existing, for the given app_entity
        if permission:
            permission_restrictions = user_role.permission_restrictions
            if permission_restrictions and permission in permission_restrictions.keys():
                permission_restriction = permission_restrictions.get(permission)
                if permission_app_entity_restriction := permission_restriction.get(
                    app_entity_type
                ):
                    return permission_app_entity_restriction

        # Verify if there is any restriction on the entity for the user
        app_entities_restrictions = user_role.app_entities_restrictions
        if app_entities_restrictions and (
            app_entity_restriction := app_entities_restrictions.get(app_entity_type)
        ):
            return app_entity_restriction

        return ALL

    @staticmethod
    def _create_or_update_user(data: UserTenantData) -> User:
        user = get_or_none(User.objects, username=data.get("username"))
        user_data = keep_keys(
            data,
            [
                "username",
                "email",
                "first_name",
                "last_name",
                "is_active",
                "is_staff",
                "is_superuser",
                "date_joined",
            ],
        )
        if user:
            update_record(user, **user_data)
            UserService._invalidate_user_cache_entries(user=user)
            return user
        else:
            user = User.objects.create(**user_data)
            post_create_idp_user.send(sender=UserService, user=user)

            return user

    @staticmethod
    def _invalidate_user_cache_entries(user: User):
        """
        Invalidate all the entries in the cache for the given user.
        To do this, find all the entries that start with the app identifier and username of the user.
        """
        if settings.IDP_USER_APP.get("USE_REDIS_CACHE", False) and not IN_DEV:
            cache.delete_pattern(f"{APP_IDENTIFIER}-{user.username}*")

    @classmethod
    def process_user(cls, data: UserRecordDict):
        """
        Extract tenants from the user record and call _update_user for each tenant.
        Remove tenant information from the payload of _update_user since it is not needed
        inside of it.

        Send signals before and after calling the _update_user method for each tenant
        separately. This gives possibility to the project to react on the user update.

        One case of handling this signal is to switch database connection to the tenant's
        database. In this way the user can be updated in the correct database.

        """
        reported_user_app_configs = UserService._get_reported_user_app_configs(data)
        tenants = reported_user_app_configs.keys()

        for tenant in tenants:
            if tenant not in TENANTS:
                logger.info(f"Tenant {tenant} not present, skipping.")
                continue

            logger.info(f"Updating user {data['username']} for tenant {tenant}")
            pre_update_idp_user.send(sender=cls.__class__, tenant=tenant)

            # Extract specific tenant information
            user_record_for_tenant = deepcopy(data)
            user_record_for_tenant["app_specific_configs"] = reported_user_app_configs[
                tenant
            ]

            try:
                with transaction.atomic(using=tenant):
                    UserService._update_user(user_record_for_tenant)  # type: ignore
            finally:
                post_update_idp_user.send(sender=cls.__class__, tenant=tenant)

    @classmethod
    def verify_if_user_exists_and_delete_roles(cls, data: UserRecordDict):
        """
        Verify if the user exists in any of the tenants and delete all the roles associated with it.
        """
        for tenant in TENANTS:
            pre_update_idp_user.send(sender=cls.__class__, tenant=tenant)

            if user := get_or_none(User.objects, username=data["username"]):
                logger.info(
                    f"Deleting roles for user {data['username']} in tenant {tenant}"
                )
                UserRole.objects.filter(user=user).delete()  # type: ignore

            post_update_idp_user.send(sender=cls.__class__, tenant=tenant)

    @staticmethod
    def _update_user(data: UserTenantData):
        """
        This method makes sure that the changes that are coming from the IDP
        for a user are propagated in the internal product Authorization Schemas

        Step 1: Create or update User Object
        Step 2: Create/Update/Delete User Roles for this user.
        """

        user = UserService._create_or_update_user(data)

        current_user_roles = defaultdict()
        for user_role in user.user_roles.all():
            current_user_roles[user_role.role] = user_role

        roles_data = data.get("app_specific_configs")

        for role, role_data in roles_data.items():
            if existing_user_role := current_user_roles.get(role):
                update_record(
                    existing_user_role,
                    permission_restrictions=role_data.get("permission_restrictions"),
                    app_entities_restrictions=role_data.get(
                        "app_entities_restrictions"
                    ),
                )
            else:
                UserRole.objects.create(
                    user=user,
                    role=role,
                    permission_restrictions=role_data.get("permission_restrictions"),
                    app_entities_restrictions=role_data.get(
                        "app_entities_restrictions"
                    ),
                )

        # Verify if any of the previous user roles is not being reported anymore
        # Delete it if this is the case
        for role, user_role in current_user_roles.items():  # type: str, UserRole
            if roles_data.get(role) is None:
                user_role.delete()

    @staticmethod
    def _get_reported_user_app_configs(data):
        return data.get("app_specific_configs", {}).get(APP_IDENTIFIER, {})

    @staticmethod
    def get_users_with_access_to_app_entity_record(
        app_entity_type: str, record_identifier: Any, roles: list[str]
    ) -> QuerySet:
        """
        Get users that have access to the required app entity record in the given roles.
        """

        assert (
            app_entity_type in APP_ENTITIES.keys()
        ), f"Unknown app entity: {app_entity_type}!"

        roles = UserRole.objects.filter(
            Q(user__is_active=True)
            & Q(role__in=roles)
            & (
                Q(app_entities_restrictions__isnull=True)
                | Q(**{f"app_entities_restrictions__{app_entity_type}__isnull": True})
                | Q(
                    **{
                        f"app_entities_restrictions__{app_entity_type}__contains": record_identifier
                    }
                )
            )
        )

        return User.objects.filter(
            pk__in=list(roles.values_list("user__pk", flat=True))
        )



import logging
from collections import defaultdict
from typing import Any, Optional, Union

from asgiref.sync import sync_to_async
from django.core.exceptions import PermissionDenied
from django.db import models
from django.db.models import Q, QuerySet
from django.http import HttpRequest

from idp_user.models import User
from idp_user.models.user_role import UserRole
from idp_user.settings import APP_ENTITIES, ROLES
from idp_user.signals import post_create_idp_user
from idp_user.utils.functions import (
    cache_user_service_results,
    get_or_none,
    keep_keys,
    parse_query_params_from_scope,
    update_record,
)
from idp_user.utils.typing import ALL, AppEntityTypeConfig, UserTenantData

logger = logging.getLogger(__name__)


class UserServiceAsync:
    # Service Methods Used by Django Application
    @staticmethod
    def get_role(request: HttpRequest):
        return request.query_params.get("role")

    @staticmethod
    async def get_role_from_scope(scope):
        query_params = parse_query_params_from_scope(scope)
        return query_params.get("role")[0] if query_params.get("role") else None

    @staticmethod
    async def authorize_and_get_records_or_get_all_allowed(
        user: User,
        role: ROLES,
        app_entity_type: str,
        app_entity_records_identifiers: Optional[list[Any]],
        permission: str = None,
    ) -> models.QuerySet:
        """
        Make sure that the user can access the entity records being requested and return them.
        If no identifiers are being provided, return all allowed entity records.

        Args:
            user:                           The user performing the request
            role:                           The role that the user is acting as.
            app_entity_type:                The app entity being accessed
            app_entity_records_identifiers: The identifiers of the records belonging to the specified app entity
            permission:                     In case of specific permissions we can have permission restrictions
                                                through IDP. The value is the name of the permission


        Raises:
            PermissionDenied: In case the requested records are not allowed
        """

        if not app_entity_records_identifiers:
            return await UserServiceAsync.get_allowed_app_entity_records(
                user=user,
                role=role,
                app_entity_type=app_entity_type,
                permission=permission,
            )
        await UserServiceAsync.authorize_app_entity_records(
            user=user,
            role=role,
            app_entity_type=app_entity_type,
            app_entity_records_identifiers=app_entity_records_identifiers,
            permission=permission,
        )
        return await UserServiceAsync._get_records(
            app_entity_type=app_entity_type,
            records_identifiers=app_entity_records_identifiers,
        )

    @staticmethod
    async def authorize_app_entity_records(
        user: User,
        role: ROLES,
        app_entity_type: str,
        app_entity_records_identifiers: list[Any],
        permission: str = None,
    ):
        """
        Verify if the user has access to the requested entity records.
        If a permission is specified and it has restrictable=True, the access is verified based on it.

        Args:
            user:                           The user performing the request
            role:                           The role that the user is acting as.
            app_entity_type:                The app entity being accessed
            app_entity_records_identifiers: The identifiers of the records belonging to the specified app entity
            permission:                     In case of specific permissions we can have permission restrictions
                                                through IDP. The value is the name of the permission

        Raises:
            PermissionDenied: In case the requested records are not allowed
        """

        assert (
            app_entity_type in APP_ENTITIES.keys()
        ), f"Unknown app entity: {app_entity_type}!"
        allowed_app_entity_records_identifiers = (
            await UserServiceAsync._get_allowed_app_entity_records_identifiers(
                user=user,
                role=role,
                app_entity_type=app_entity_type,
                permission=permission,
            )
        )
        if allowed_app_entity_records_identifiers == ALL:
            return

        if not set(app_entity_records_identifiers).issubset(
            set(allowed_app_entity_records_identifiers)
        ):
            raise PermissionDenied(
                "You are not allowed to access the records in the requested entity!"
            )

    @staticmethod
    async def get_allowed_app_entity_records(
        user: User, role: ROLES, app_entity_type: str, permission: str = None
    ) -> models.QuerySet:
        """
        Gets the app entity records that the user can access.

        Args:
            user:               The user performing the request
            role:               The role that the user is acting as.
            app_entity_type:    The app entity being accessed
            permission:         In case of specific permissions we can have permission restrictions
                                    through IDP. The value is the name of the permission

        Returns:
            QuerySet of App Entity Records that the user can access
        """

        assert (
            app_entity_type in APP_ENTITIES.keys()
        ), f"Unknown app entity: {app_entity_type}!"

        allowed_app_entity_records_identifiers = (
            await UserServiceAsync._get_allowed_app_entity_records_identifiers(
                user=user,
                role=role,
                app_entity_type=app_entity_type,
                permission=permission,
            )
        )
        return await UserServiceAsync._get_records(
            app_entity_type=app_entity_type,
            records_identifiers=allowed_app_entity_records_identifiers,
        )

    @staticmethod
    async def _get_app_entity_type_configs(app_entity_type: str) -> AppEntityTypeConfig:
        try:
            return APP_ENTITIES[app_entity_type]
        except KeyError as e:
            raise KeyError(
                f"No config declared for app_entity_type={app_entity_type} "
                f"in IDP_USER_APP['APP_ENTITIES']!"
            ) from e

    @staticmethod
    async def _get_records(
        app_entity_type: str, records_identifiers: Union[list[Any], ALL]
    ) -> models.QuerySet:
        app_entity_type_configs = await UserServiceAsync._get_app_entity_type_configs(
            app_entity_type
        )
        model = app_entity_type_configs["model"]

        if records_identifiers == ALL:
            return await sync_to_async(list)(model.objects.all())
        model_identifier_attr = app_entity_type_configs["identifier_attr"]
        return await sync_to_async(model.objects.filter)(
            **{f"{model_identifier_attr}__in": records_identifiers}
        )

    @staticmethod
    @cache_user_service_results
    async def _get_allowed_app_entity_records_identifiers(
        user: User, role: ROLES, app_entity_type: str, permission: str = None
    ) -> Union[list[Any], ALL]:
        """
        Gets the identifiers of the app entity records that the user can access.
        If no restriction both on role and permission level, return '__all__'

        Args:
            user:               The user performing the request
            role:               The role that the user is acting as.
            app_entity_type:    The app entity being accessed
            permission:         In case of specific permissions we can have permission restrictions
                                    through IDP. The value is the name of the permission

        Returns:
            List of identifiers of App Entity Records that the user can access
        """
        assert ROLES.as_dict().get(role) is not None, f"Role does not exist: {role}"

        try:
            user_role = await UserRole.objects.aget(user=user, role=role)
        except UserRole.DoesNotExist:
            return []

        # Permission restriction get precedence, if existing, for the given app_entity
        if permission:
            permission_restrictions = user_role.permission_restrictions
            if permission_restrictions and permission in permission_restrictions.keys():
                permission_restriction = permission_restrictions.get(permission)
                if permission_app_entity_restriction := permission_restriction.get(
                    app_entity_type
                ):
                    return permission_app_entity_restriction

        # Verify if there is any restriction on the entity for the user
        app_entities_restrictions = user_role.app_entities_restrictions
        if app_entities_restrictions and (
            app_entity_restriction := app_entities_restrictions.get(app_entity_type)
        ):
            return app_entity_restriction

        return ALL

    @staticmethod
    async def _create_or_update_user(data: UserTenantData) -> User:
        user = await sync_to_async(get_or_none)(
            User.objects, username=data.get("username")
        )
        user_data = keep_keys(
            data,
            [
                "username",
                "email",
                "first_name",
                "last_name",
                "is_active",
                "is_staff",
                "is_superuser",
                "date_joined",
            ],
        )
        if user:
            await sync_to_async(update_record)(user, **user_data)
        else:
            user = await User.objects.acreate(**user_data)
            post_create_idp_user.send(sender=UserServiceAsync, user=user)
        return user

    @staticmethod
    async def _update_user(data: UserTenantData):
        """
        This method makes sure that the changes that are coming from the IDP
        for a user are propagated in the internal product Authorization Schemas

        Step 1: Create or update User Object
        Step 2: Create/Update/Delete User Roles for this user.
        """

        user = await UserServiceAsync._create_or_update_user(data)

        current_user_roles = defaultdict()
        async for user_role in user.user_roles.all():
            current_user_roles[user_role.role] = user_role

        roles_data = data.get("app_specific_configs")

        for role, role_data in roles_data.items():
            if existing_user_role := current_user_roles.get(role):
                await sync_to_async(update_record)(
                    existing_user_role,
                    permission_restrictions=role_data.get("permission_restrictions"),
                    app_entities_restrictions=role_data.get(
                        "app_entities_restrictions"
                    ),
                )
            else:
                await UserRole.objects.acreate(
                    user=user,
                    role=role,
                    permission_restrictions=role_data.get("permission_restrictions"),
                    app_entities_restrictions=role_data.get(
                        "app_entities_restrictions"
                    ),
                )

        # Verify if any of the previous user roles is not being reported anymore
        # Delete it if this is the case
        for role, user_role in current_user_roles.items():  # type: str, UserRole
            if roles_data.get(role) is None:
                user_role.delete()

    @staticmethod
    async def get_users_with_access_to_app_entity_record(
        app_entity_type: str, record_identifier: Any, roles: list[str]
    ) -> QuerySet:
        """
        Get users that have access to the required app entity record in the given roles.
        """

        assert (
            app_entity_type in APP_ENTITIES.keys()
        ), f"Unknown app entity: {app_entity_type}!"

        roles = await sync_to_async(UserRole.objects.filter)(
            Q(user__is_active=True)
            & Q(role__in=roles)
            & (
                Q(app_entities_restrictions__isnull=True)
                | Q(**{f"app_entities_restrictions__{app_entity_type}__isnull": True})
                | Q(
                    **{
                        f"app_entities_restrictions__{app_entity_type}__contains": record_identifier
                    }
                )
            )
        )
        return await sync_to_async(User.objects.filter)(
            pk__in=list(roles.values_list("user__pk", flat=True))
        )

    @staticmethod
    async def get_user(username: str) -> User:
        """
        Get user by username
        """
        return await sync_to_async(get_or_none)(User.objects, username=username)



import logging
from datetime import datetime
from typing import Any, Type

from django.db import models

from idp_user.producer import Producer
from idp_user.settings import (
    APP_ENTITIES,
    APP_ENTITY_RECORD_EVENT_TOPIC,
    APP_IDENTIFIER,
)
from idp_user.utils.exceptions import UnsupportedAppEntityType
from idp_user.utils.typing import AppEntityRecordEventDict

logger = logging.getLogger(__name__)


class BaseUserService:
    @classmethod
    def send_app_entity_record_event_to_kafka(
        cls, app_entity_type: str, app_entity_record: Any, deleted=False
    ):
        app_entity_type_config = APP_ENTITIES[app_entity_type]

        event: AppEntityRecordEventDict = {
            "app_identifier": APP_IDENTIFIER,
            "app_entity_type": app_entity_type,
            "record_identifier": getattr(
                app_entity_record, app_entity_type_config["identifier_attr"]
            ),
            "label": getattr(app_entity_record, app_entity_type_config["label_attr"]),
            "deleted": deleted,
        }

        logger.info(f"Sending update {event}...")

        Producer().send_message(
            topic=APP_ENTITY_RECORD_EVENT_TOPIC, key=str(datetime.now()), data=event
        )

    @classmethod
    def _get_app_entity_type_from_model(cls, model: Type[models.Model]):
        for (
            app_entity_type,
            config,
        ) in APP_ENTITIES.items():  # type: str, AppEntityTypeConfig
            if config["model"] == model:
                return app_entity_type

        raise UnsupportedAppEntityType(model)

    @classmethod
    def process_app_entity_record_post_save(
        cls, sender: Type[models.Model], instance, **kwargs
    ):
        """
        Whenever an app entity record is saved (created/updated),
        send a message to Kafka to notify the IDP.

        kwargs are required for signal receivers.
        """

        cls.send_app_entity_record_event_to_kafka(
            app_entity_type=cls._get_app_entity_type_from_model(sender),
            app_entity_record=instance,
        )

    @classmethod
    def process_app_entity_record_post_delete(
        cls, sender: Type[models.Model], instance, **kwargs
    ):
        """
        Whenever an app entity record is deleted,
        send a message to Kafka to notify the IDP.

        kwargs are required for signal receivers.
        """

        cls.send_app_entity_record_event_to_kafka(
            app_entity_type=cls._get_app_entity_type_from_model(sender),
            app_entity_record=instance,
            deleted=True,
        )



class AuthenticationError(Exception):
    def __init__(self, detail: str = "Authentication Error"):
        super().__init__(detail)


class MissingHeaderError(Exception):
    def __init__(self, header_name):
        super().__init__(f"Missing header: {header_name}")


class UnsupportedAppEntityType(Exception):
    def __init__(self, app_entity_type):
        super().__init__(f"Unsupported app entity type: {str(app_entity_type)}")



from django.conf import settings
from django.utils.module_loading import import_string

ASYNC_MODE = settings.IDP_USER_APP.get("ASYNC_MODE", False)
APP_IDENTIFIER = settings.IDP_USER_APP["APP_IDENTIFIER"]
IN_DEV = settings.APP_ENV == "development"
ROLES = import_string(settings.IDP_USER_APP.get("ROLES"))
APP_ENTITIES = settings.IDP_USER_APP.get("APP_ENTITIES") or {}
TENANTS = settings.IDP_USER_APP.get("TENANTS") or list(settings.DATABASES.keys())
CONSUMER_APP_ENV = settings.IDP_USER_APP.get("CONSUMER_APP_ENV") or settings.APP_ENV

if APP_ENTITIES:
    for _, config_dict in APP_ENTITIES.items():
        config_dict["model"] = import_string(config_dict["model"])


APP_ENTITY_RECORD_EVENT_TOPIC = f"{CONSUMER_APP_ENV}_app_entity_record_events"

AWS_S3_REGION_NAME = getattr(settings, "AWS_S3_REGION_NAME", None) or "eu-central-1"



import aiohttp
from django.conf import settings

from idp_user.services.async_user import UserServiceAsync
from idp_user.utils.functions import parse_query_params_from_scope

IDP_URL = settings.IDP_USER_APP.get("IDP_URL")


class IDPChannelsAuthenticationMiddleware:
    user_service_cls = UserServiceAsync

    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        if scope.get("user"):
            return await self.app(scope, receive, send)

        query_params = parse_query_params_from_scope(scope)
        authorization_access_token = query_params.get("authorization", [None])[0]
        if not authorization_access_token:
            return await self.app(scope, receive, send)
        username = await self._get_username(authorization_access_token)
        if not username:
            return await self.app(scope, receive, send)
        user = await self._get_user(username)
        if user:
            scope["user"] = user
            scope["authorization_access_token"] = authorization_access_token

        return await self.app(scope, receive, send)

    async def _get_username(self, access_token):
        headers = {"Authorization": f"Bearer {access_token}"}
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.get(f"{IDP_URL}/api/users/me/") as resp:
                if resp.status != 200:
                    return None
                response = await resp.json()
                return response.get("username")

    async def _get_user(self, username):
        return await self.user_service_cls.get_user(username=username)



import logging
from typing import Callable, List, Literal, Optional

import jwt
from django.conf import settings
from rest_framework import authentication
from rest_framework.request import Request

from idp_user.models.user import User
from idp_user.utils.exceptions import AuthenticationError, MissingHeaderError
from idp_user.utils.functions import get_or_none
from idp_user.utils.typing import JwtData

logger = logging.getLogger(__name__)

APP_IDENTIFIER = settings.IDP_USER_APP.get("APP_IDENTIFIER")
IDP_URL = settings.IDP_USER_APP.get("IDP_URL")


class AuthenticationBackend(authentication.TokenAuthentication):
    @classmethod
    def _access_token_required(cls, request: Request):
        try:
            cls._header_required(
                request, "Authorization", "token", lambda x: x.startswith("Bearer ")
            )
        except MissingHeaderError as e:
            raise AuthenticationError("Missing Token.") from e

    @classmethod
    def _header_required(
        cls,
        request: Request,
        header: str,
        header_human_name: str,
        validation_func: Optional[Callable] = None,
    ):
        header = cls._get_request_header(request, header)
        if not header:
            raise MissingHeaderError(header_human_name)
        if validation_func and not validation_func(header):
            raise MissingHeaderError(header_human_name)

    @classmethod
    def _get_access_token(cls, request: Request) -> str:
        auth_header = cls._get_request_header(request, "Authorization")
        return auth_header.replace("Bearer ", "")

    @classmethod
    def _get_jwt_payload(cls, request: Request) -> JwtData:
        try:
            # Signature already verified from ingress before
            data = jwt.decode(
                cls._get_access_token(request),
                algorithms=["HS256"],
                options={"verify_signature": False},
            )
        except Exception as e:
            raise AuthenticationError("Invalid token.") from e
        return data

    @classmethod
    def _verify_jwt_claims(cls, jwt_data: JwtData, claims: List[Literal["username"]]):
        for claim in claims:
            cls._verify_jwt_claim(jwt_data, claim)

    @classmethod
    def _verify_jwt_claim(cls, jwt_data: JwtData, claim: Literal["username"]):
        try:
            jwt_data[claim]
        except KeyError as e:
            raise AuthenticationError("Invalid token.") from e

    @staticmethod
    def _get_request_header(request: Request, header: str) -> str:
        return request.headers.get(header) or request.META.get(header)

    @staticmethod
    def _get_user(username: str) -> Optional[User]:
        return get_or_none(User.objects, username=username)

    def authenticate(self, request: Request):
        try:
            self._access_token_required(request)
            jwt_data = self._get_jwt_payload(request)
            self._verify_jwt_claims(jwt_data, ["username"])
            user = self._get_user(jwt_data["username"])
            auth = self
        except (AuthenticationError, MissingHeaderError):
            user = None
            auth = None
        return user, auth



from django.db import models


class UserRole(models.Model):
    user = models.ForeignKey(
        to="idp_user.User", related_name="user_roles", on_delete=models.CASCADE
    )
    role = models.CharField(max_length=140)
    # This dictionary contains explicit restrictions about the app entities that the user can access, in the form:
    # {<entity_type>: [1, 2]}
    app_entities_restrictions = models.JSONField(null=True)
    # This dictionary contains explicit restrictions regarding app permissions, in the form:
    # {"perform_operation_1": {<entity_type>: [1, 2]}, "perform_operation_2": false}
    permission_restrictions = models.JSONField(default=dict)

    class Meta:
        unique_together = [("user", "role")]



from django.conf import settings
from django.core.checks import Error, Warning, register


@register()
def check_idp_user_settings(*args, **kwargs):
    def verify_idp_user_app_attr_exists(_attr, error=False):
        if _attr not in idp_user_settings:
            issue_kwargs = {
                "msg": f"Missing {_attr} in IDP_USER_APP.",
                "hint": f"You must declare the attribute {_attr} in the dict IDP_USER_APP in settings.py.",
                "obj": settings,
                "id": f"idp_user.{_attr}",
            }
            if error:
                issues.append(Error(**issue_kwargs))
            else:
                issues.append(Warning(**issue_kwargs))

    issues = []

    try:
        if hasattr(settings, "APP_ENV") and settings.APP_ENV not in [
            "development",
            "staging",
            "production",
            "demo",
        ]:
            issues.append(
                Error(
                    "Wrong APP_ENV.",
                    hint="You must declare the APP_ENV (development/staging/production/demo) variable in settings.py.",
                    obj=settings,
                    id="idp_user.E000",
                )
            )
    except AttributeError:
        issues.append(
            Error(
                "Missing APP_ENV.",
                hint="You must declare the APP_ENV (development/staging/production) variable in settings.py.",
                obj=settings,
                id="idp_user.E001",
            )
        )

    try:
        auth_user_model = settings.AUTH_USER_MODEL
        if auth_user_model != "idp_user.User":
            issues.append(
                Warning(
                    "Wrong AUTH_USER_MODEL.",
                    hint="You must set AUTH_USER_MODEL = 'idp_user.User' in settings.py.",
                    obj=settings,
                    id="idp_user.E002",
                )
            )

    except AttributeError:
        issues.append(
            Warning(
                "Missing AUTH_USER_MODEL",
                hint="You must set AUTH_USER_MODEL = 'idp_user.user' in settings.py.",
                obj=settings,
                id="idp_user.E003",
            )
        )

    try:
        idp_user_settings = settings.IDP_USER_APP

        for attr in ["APP_IDENTIFIER", "FAUST_APP_PATH"]:
            verify_idp_user_app_attr_exists(attr, error=True)

        for attr in ["ROLES"]:
            verify_idp_user_app_attr_exists(attr, error=False)

    except AttributeError:
        issues.append(
            Error(
                "Missing IDP_USER_APP.",
                hint="You must declare the idp_user settings in the variable IDP_USER_APP in settings.py.",
                obj=settings,
                id="idp_user.E004",
            )
        )

    return issues



import json

from aiokafka import AIOKafkaProducer
from django.core.serializers.json import DjangoJSONEncoder
from kafka import KafkaProducer

from idp_user.utils.classes import Singleton
from idp_user.utils.functions import get_kafka_bootstrap_servers
import ssl


class Producer(metaclass=Singleton):
    __connection = None

    def __init__(self):
        self.__connection = KafkaProducer(
            bootstrap_servers=get_kafka_bootstrap_servers(include_uri_scheme=False),
            value_serializer=lambda v: json.dumps(v, cls=DjangoJSONEncoder).encode(
                "utf-8"
            ),
            api_version=(2, 6, 2),
            ssl_context=ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH),
            security_protocol="SSL"
        )

    def send_message(self, topic: str, key: str, data: dict):
        self.__connection.send(topic=topic, key=key.encode("utf-8"), value=data)
        # Sometimes messages do not get sent.
        # Flushing after each message seems to solve the issue
        self.__connection.flush()


class AioKafkaProducer(metaclass=Singleton):
    _producer = None

    async def get_producer(self):
        if self._producer is None:
            self._producer = AIOKafkaProducer(
                bootstrap_servers=get_kafka_bootstrap_servers(include_uri_scheme=False),
                value_serializer=lambda v: json.dumps(v, cls=DjangoJSONEncoder).encode(
                    "utf-8"
                ),
            )
            await self._producer.start()
        return self._producer

    async def send_message(self, topic: str, key: str, data: dict):
        producer = await self.get_producer()
        await producer.send_and_wait(topic=topic, key=key.encode("utf-8"), value=data)

    async def close(self):
        await self._producer.stop()
